// oop_ex119.cpp
// virtual bases and copy constructors

#include <iostream>

using namespace std;

//-----X-----------------

class X
{
	int x;
public:
	X(int xx) : x(xx) { }

	//== copy constructor ==
	X(const X& other) : x(other.x) { } 

	virtual void print()
	{
		cout << "The value of x of X sub-object: " << x << endl;
	}
};

//-----Y-----------------

//class Y : public X
class Y : virtual public X
{
	int y;
public:
	Y(int xx, int yy) : X(xx), y(yy) { }	 //X(xx) will be ignored when invoked by the constructors of Z

	//== copy constructor ==
	Y(const Y& other) : X(other), y(other.y) { } //X(other) will be ignored when invoked by the copy constructor of Z

	virtual void print()
	{
		X::print();
		cout << "The value of y of Y sub-object: " << y << endl;
	}
};


//-----T-----------------

//class T : public X
class T : virtual public X
{
	int t;
public:
	T(int xx, int tt) : X(xx), t(tt) { } //X(xx) will be ignored when invoked by the constructors of U

	//== copy constructor ==
	T(const T& other) : X(other), t(other.t) { }	//X(other) will be ignored when invoked by the copy constructor of U

	virtual void print()
	{
		X::print();
		cout << "The value of t of T sub-object: " << t << endl;
	}
};

//------Z-----------------

class Z : public Y
{
	int z;
public:
	Z(int xx, int yy, int zz) : Y(xx, yy), X(xx), z(zz) { }

	//== copy constructor ==
	Z(const Z& other) : Y(other), X(other), z(other.z) { } //X(other) must be invoked explicitly

	virtual void print()
	{
		Y::print();
		cout << "The value of z of Z sub-object: " << z << endl;
	}
};

//-----W-----------------

class W : public Z
{
	int w;
public:
	W(int xx, int yy, int zz, int ww) : Z(xx, yy, zz), X(xx), w(ww) { }

	//== copy constructor ==
	W(const W& other) : Z(other), X(other), w(other.w) { } //X(other) must be invoked explicitly

	virtual void print()
	{
		Z::print();
		cout << "The value of w of W sub-object: " << w << endl;
	}
};

//------U-----------------

class U : public W, public T
{
	int u;
public:
	U(int xx, int yy, int zz, int ww, int tt, int uu)
		: W(xx, yy, zz, ww), T(xx, tt), X(xx), u(uu) { }

	//== copy constructor ==
	U(const U& other) : W(other), T(other), X(other), u(other.u) { } //X(other) must be invoked explicitly

	virtual void print()
	{
		X::print();	//this is OK now!
		W::print();
		T::print();
		cout << "The value of u of U sub-object: " << u << endl;
	}
};

//-----main()--------------

void main()
{
	cout << "W object coming up: " << endl;
	W wobj(1111, 2222, 3333, 4444);
	wobj.print();
	cout << endl;

	cout << "W's duplicate object coming up: " << endl;
	W wobj_2 = wobj;
	wobj_2.print();
	cout << endl;

	cout << "U object coming up: " << endl;
	U uobj(111111, 222222, 333333, 444444, 555555, 666666);
	uobj.print();
	cout << endl;

	cout << "U's duplicate object coming up: " << endl;
	U uobj_2 = uobj;
	uobj_2.print();
	cout << endl;

	system("pause");
}