// oop_ex120.cpp
// virtual bases and copy assignment operators

#include <iostream>

using namespace std;

//-----X-----------------

class X
{
	int x;
public:
	X(int xx) : x(xx) { }

	X(const X& other) : x(other.x) { }

	//== copy assignment operator ==
	X& operator=(const X& other) 
	{
		//cout<<"calling X& operator=(const X& other)"<<endl;

		if (this != &other)
			x = other.x;

		return *this;
	}

	virtual void print()
	{
		cout << "The value of x of X sub-object: " << x << endl;
	}
};

//-----Y-----------------

//class Y : public X
class Y : virtual public X
{
	int y;
public:
	Y(int xx, int yy) : X(xx), y(yy) { }

	Y(const Y& other) : X(other), y(other.y) { }

	//== copy assignment operator ==
	Y& operator=(const Y& other) 
	{
		if (this != &other)
		{
			X::operator=(other); //<= invoke copy assignment operator of X
			y = other.y;
		}

		return *this;
	}

	virtual void print()
	{
		X::print();
		cout << "The value of y of Y sub-object: " << y << endl;
	}
};


//-----T-----------------

//class T : public X
class T : virtual public X
{
	int t;
public:
	T(int xx, int tt) : X(xx), t(tt) { }

	T(const T& other) : X(other), t(other.t) { }

	//== copy assignment operator ==
	T& operator=(const T& other) 
	{	
		// ~~using another syntex~~

		if (this == &other) 
			return *this;

		X::operator=(other); //<= invoke copy assignment operator of X
		t = other.t;
		return *this;
	}

	virtual void print()
	{
		X::print();
		cout << "The value of t of T sub-object: " << t << endl;
	}
};

//-----Z-----------------

class Z : public Y
{
	int z;
public:
	Z(int xx, int yy, int zz) : Y(xx, yy), X(xx), z(zz) { }

	Z(const Z& other) : Y(other), X(other), z(other.z) { }

	//== copy assignment operator ==
	Z& operator=(const Z& other) 
	{
		if (this != &other)
		{
			Y::operator=(other);	//<= invoke copy assignment operator of Y
			z = other.z;
		}

		return *this;
	}

	virtual void print()
	{
		Y::print();
		cout << "The value of z of Z sub-object: " << z << endl;
	}
};

//-----W-----------------

class W : public Z
{
	int w;
public:
	W(int xx, int yy, int zz, int ww) : Z(xx, yy, zz), X(xx), w(ww) { }

	W(const W& other) : Z(other), X(other), w(other.w) { }

	//== copy assignment operator ==
	W& operator=(const W& other) 
	{
		if (this != &other)
		{
			Z::operator=(other);	//<= invoke copy assignment operator of Z
			w = other.w;
		}

		return *this;
	}

	virtual void print()
	{
		Z::print();
		cout << "The value of w of W sub-object: " << w << endl;
	}
};

//-----U-----------------

class U : public W, public T
{
	int u;
public:
	U(int xx, int yy, int zz, int ww, int tt, int uu)
		: W(xx, yy, zz, ww), T(xx, tt), X(xx), u(uu) { }

	U(const U& other) : W(other), T(other), X(other), u(other.u) { }

	//== copy assignment operator == 
	U& operator=(const U& other) 
	{
		if (this != &other)
		{
			W::operator=(other); //<= invoke copy assignment operator of W
			T::operator=(other);	 //<= invoke copy assignment operator of T
			u = other.u;
		}

		return *this;
	}

	virtual void print()
	{
		//X::print();	
		W::print();
		T::print();
		cout << "The value of u of U sub-object: " << u << endl;
	}
};

//-----main()--------------

void main()
{
	cout << "U object coming up: " << endl;
	U uobj(111111, 222222, 333333, 444444, 555555, 666666);
	uobj.print();
	cout << endl;

	U uobj_3(1, 2, 3, 4, 5, 6);
	uobj_3.print();

	uobj_3 = uobj; //<= call copy assignment operator of U

	uobj_3.print();

	system("pause");
}